from collections import OrderedDict

from os_config import Config
from sanic import Blueprint
from sanic.views import HTTPMethodView

from os_sanic.utils import load_class


class View(object):
    def __init__(self, uri, view_cls, config):
        self.uri = uri
        self.view_cls = view_cls
        self.config = config

    @classmethod
    def create(cls, application, blueprint, view_cfg, user_config):

        uri, view_class = view_cfg.uri, view_cfg.view_class
        config = Config.create()
        Config.update(config, view_cfg)
        Config.update(config, user_config)
        Config.pop(config, 'uri')
        Config.pop(config, 'view_class')

        package = None
        if view_class.startswith('.'):
            package = application.app_cfg.package
        view_cls = load_class(
            view_class, HTTPMethodView, package=package)

        kwargs = {}
        if len(config) > 0:
            kwargs['config'] = config

        blueprint.add_route(view_cls.as_view(**kwargs), uri)

        return cls(uri, view_cls, config)


class ViewManager(object):

    def __init__(self, application, blueprint):
        self.application = application
        self.blueprint = blueprint
        self._views = OrderedDict()
        self.logger = application.get_logger(self.__class__.__name__)

    @property
    def views(self):
        return self._views.values()

    def _load_view(self, view_cfg, user_config):

        if isinstance(view_cfg, tuple):
            view_cfg = Config.create(uri=view_cfg[0], view_class=view_cfg[1])

        uri, view_class = view_cfg.uri, view_cfg.view_class

        if uri in self._views:
            self.logger.warn(f'View already existed, {uri} {view_class}')
            return

        view = View.create(self.application, self.blueprint,
                           view_cfg, user_config)
        self._views[uri] = view
        pth = self.blueprint.url_prefix+uri if self.blueprint.url_prefix else uri
        self.logger.debug(f'Load view, {pth} {view.view_cls}')

    def load_view(self, view_cfg, user_config):
        try:
            self._load_view(view_cfg, user_config)
        except Exception as e:
            self.logger.error(f'Load view error, {e}')

    @classmethod
    def create(cls, application):
        prefix = None
        if not Config.get(application.app_cfg, 'root'):
            prefix = Config.get(application.app_cfg,
                                'prefix', '/' + application.name)
        blueprint = Blueprint(application.name, url_prefix=prefix)

        view_manager = cls(application, blueprint)

        configs = {}
        for v in Config.get(application.user_config, 'VIEWS', []):
            uri = Config.get(v, 'uri')
            if uri:
                Config.pop(v, 'uri')
                configs[uri] = v

        for view_cfg in Config.get(application.core_config, 'VIEWS', []):
            view_manager.load_view(view_cfg, configs.get(
                view_cfg.uri, Config.create()))

        application.sanic.blueprint(blueprint)

        return view_manager



import os
import inspect
from collections import OrderedDict
from functools import partial
from importlib import import_module
from inspect import isawaitable

from os_config import Config

from os_sanic.extension import ExtensionManager
from os_sanic.log import getLogger
from os_sanic.view import ViewManager
from os_sanic.workflow import Workflowable


class Application(Workflowable):

    def __init__(self, sanic, app_cfg, core_config, user_config):
        self.sanic = sanic
        self.app_cfg = app_cfg
        self.core_config = core_config
        self.user_config = user_config

        self.logger = getLogger(f'App.{app_cfg.name}')

        self.extension_manager = ExtensionManager.create(self)

        self.view_manager = ViewManager.create(self)

        [setattr(self, method, partial(self.__call, method))
         for method in ('run', 'setup', 'cleanup')]

    @property
    def name(self):
        return self.app_cfg.name

    def get_logger(self, tag):
        return getLogger(f'App.{self.name}.{tag}')

    @property
    def package(self):
        return self.app_cfg.package

    async def __call(self, method):
        try:
            await getattr(self.extension_manager, method)()
        except Exception as e:
            self.logger.error(f'Method error {method}, {e}')

    @staticmethod
    def create(sanic, app_cfg):

        app_module = import_module('.'.join((app_cfg.package, 'app')))
        core_config = Config.from_object(app_module)
        user_config = Config.create()
        if Config.get(app_cfg, 'config'):
            f = os.path.join(os.getcwd(), app_cfg.config)
            user_config = Config.from_pyfile(f)

        return Application(sanic, app_cfg, core_config, user_config)


class ApplicationManager(Workflowable):

    def __init__(self, sanic):
        self.sanic = sanic
        self.logger = getLogger(self.__class__.__name__)
        self._apps = OrderedDict()
        self._root_app = None
        [setattr(self, method, partial(self.__call, method))
         for method in ('run', 'setup', 'cleanup')]

    async def __call(self, method):
        iter = self._apps.keys()
        if method == 'cleanup':
            iter = sorted(iter, reverse=True)

        for key in iter:
            app = self._apps[key]
            try:
                await getattr(app, method)()
            except Exception as e:
                self.logger.error(f'Application error {key}.{method}, {e}')

    def get_extension(self, extension_path):
        app_name, extension_name = extension_path.split('.')
        app = self.get_app(app_name)
        return app.extension_manager.get_extension(extension_name)

    def get_app(self, name):
        return self._apps[name]

    @property
    def apps(self):
        return self._apps.values()

    def load_app(self, app_cfg):
        try:
            if isinstance(app_cfg, str):
                app_cfg = Config.create(package=app_cfg)
            if not Config.get(app_cfg, 'package'):
                self.logger.warn(f'Load app skip, no package {app_cfg}')
                return

            if Config.get(app_cfg, 'root') and self._root_app:
                self.logger.warn(f'Root app exist: {self._root_app.name}')
                Config.pop(app_cfg, 'root')

            app_name = app_cfg.package.split('.')[-1]
            app_name = Config.get(app_cfg, 'name', app_name)
            app_cfg.name = app_name
            if app_name in self._apps:
                self.logger.warn(f'App already exists, {app_name}')
                return

            self.logger.debug(
                f'Load app, {app_name} <package \'{app_cfg.package}>\'')
            app = Application.create(self.sanic, app_cfg)
            if Config.get(app_cfg, 'root'):
                self._root_app = app
            self._apps[app_name] = app
        except Exception as e:
            self.logger.error(f'Load app fail, {e}, {app_cfg}')

    @classmethod
    def create(cls, sanic):

        assert not hasattr(
            sanic, 'application'), 'sanic instance already has \'application\' attribute'

        am = cls(sanic)
        for app_cfg in Config.create(
                apps=sanic.config.get('INSTALLED_APPS', [])).apps:
            am.load_app(app_cfg)
        sanic.application = am
        return am
